// Output from p2c 1.21alpha-07.Dec.93, the Pascal-to-C translator
// $Id: bm3.cpp,v 1.2 2004/09/01 12:45:28 micheli Exp $
// From input file "bm3.pas"

#include <bm3.h>
#include <common.h>
#include <mathe1.h>
#include <mathe2.h>
#include <mathe4.h>
#include <fourier.h>
#include <random.h>
#include <arma.h>
#include <sfu.h>

#include <cstring>
#include <cmath>
#include <iostream>
#include <ctime>

//! value for the random number generator
extern long randseed;

/*! initialise common values from the command line */
static void
init( LOC_run& V )
{
    TRACE("init(.)");
    // Ctes for this application
    V.oscamp = 3.0;
    V.oscper = 10.0;
    V.vdnfac = 0.2;
    V.vdnper = 1.0;
    V.drofac = 0.8;
    V.droper = 3.1;
    V.del0   = 1e-3;

	V.prim_freq = get_prim_freq(V.spect_type);
	V.p = getp(V.spect_type);
	V.q = getq(V.spect_type);

	V.dotn = (10 - (V.sig_group & 1) * 9.0) * 100;
	V.sqrtnoise = sqrt(0.2) * ((V.sig_group & 3) / 2);
	V.del = V.del0 * ((V.sig_group & 7) / 4);
	V.vosc = (bool)((V.sig_group & 15) / 8);
	V.vvdn = (bool)((V.sig_group & 31) / 16);
	V.vdro = (bool)((V.sig_group & 63) / 32);
	V.bias = (V.bias_type > 0);
	V.mean = (byte)(V.bias_type == 2) * 2.0;
//    V.vvar = 1.0;
	V.quer = (byte)(V.bias_type == 3);
}

static void
ngs(LOC_run& V)
{
    TRACE("ngs(.)");
	V.gsto = V.gst;
	V.gsuo = V.gsu;
	V.gsu = 0.0;
	size_t pp = V.p + 1;
	size_t qq = V.q + 1;

	V.ee[V.gsc % qq] = gaussran() * V.ees;
    
	int FORLIM = V.q;
	for (int i = 0; i <= FORLIM; ++i) {
		V.gsu += V.bb[i] * V.ee[(V.gsc + qq - i) % qq];
	}
    
	FORLIM = V.p;
	for (int i = 1; i <= FORLIM; ++i) {
		V.gsu -= V.aa[i] * V.yy[(V.gsc + pp - i) % pp];
	}
    
	V.yy[V.gsc % pp] = V.gsu;

	V.gsu += V.mean + (byte)V.vosc * V.oscamp *
						cos(2 * M_PI * V.gsc / V.prim_freq / V.oscper + V.pper);
	V.gst = V.gsc / V.prim_freq;
	++V.gsc;
    
	double TEMP = V.quer;
	V.gss += sqrt(fabs(TEMP * TEMP + V.gsuo * V.gsuo)) / V.prim_freq;
//	if (prim) {
//		fprintf(V.psf, "% .5E % .5E\n", V.gst, V.gsu);
//		meldung(mind((double)V.gsc / psn, 1.0), false, V);
//	}
}

static void
nts(LOC_run& V)
{
    TRACE("nts(.)");
	double DUMMY, TEMP;

	V.tsto = V.tst;
	V.tsuo = V.tsu;
	do {
		if (V.tsx >= 1) {
            
            // USE NGS
		    ngs(V);
            
		    V.vv = 1.0;
		    if (V.vvdn) {
    		    V.vv *= 1 + (1 - V.vdnfac) / (1 + V.vdnfac) *
						    cos(2 * M_PI * V.gst / V.vdnper);
		    }
		    if (V.vdro) {
	    	    V.vv = V.vv * (byte)(modf(V.gst / V.droper, &DUMMY) <
									      V.drofac) / V.drofac;
		    }
		    V.tsx = 0.0;
		}
        
		if (V.bias) {
		    if (V.vv > 0 &&
			    sqrt(fabs((TEMP = V.quer, TEMP * TEMP) + V.gsu * V.gsu)) > 0) {
    		    double TEMP1 = V.quer;
	    	    V.tsx += V.dsm * V.prim_freq / V.vv / sqrt(
					       fabs(TEMP1 * TEMP1 + V.gsu * V.gsu)) * expran();
		    } else {
		        V.tsx = 1.0;
		    }
		} else {
		    if (V.vv > 0) {
    		    V.tsx += V.dtm * V.prim_freq / V.vv * expran();
		    } else {
	    	    V.tsx = 1.0;
		    }
		}
	} while (V.tsx >= 1);
    
	V.tst = V.gst + V.tsx / V.prim_freq;
	V.tsu = V.gsu;
	++V.tsc;
}

static void
nps(LOC_run& V)
{
    TRACE("nps(.)");
    double a;

	do {
        // USE NTS
		nts(V);
        
		V.psto = V.pst;
		V.psuo = V.psu;
		V.pstto = V.pstt;
		V.psvo = V.psv;

		do {
		    double TEMP = gleichran();
		    double TEMP1 = gleichran();
		    a = 1 - TEMP * TEMP - TEMP1 * TEMP1;
		    if (a > 0) {
		        TEMP = V.quer;
		        V.pstt = V.del0 * V.dsm / V.dtm / sqrt(
					           fabs(TEMP * TEMP + V.tsu * V.tsu)) * sqrt(a);
		    }
		} while (a <= 0);
        
		V.pst = V.tst;
		V.psu = V.tsu + V.sqrtnoise * gaussran();
		V.psv = true;
		if (V.del > 0 && V.psto + V.pstto / 2 > V.pst - V.pstt / 2) {
    		V.psvo = false;
	    	V.psv = false;
		}
	} while (V.psvo != true);
    
	fprintf(V.psf, "% .5E % .5E % .5E\n", V.psto, V.psuo, V.pstto);
	++V.psc;
	meldung(mind((double)V.psc / V.psn, 1.0), false, V);
}

static void
run(LOC_run& V)
{
    TRACE("run()");
    double yys, yyv, eev, rxx0;
    tcomplex* mfourier_;
    int i;
    double s1 = 0.0;
    double* cc, *cn, *co, *ryy, *rey;
    int FORLIM;
    size_t pp = V.p + 1;
	size_t qq = V.q + 1;
    
    V.psf = NULL;
    V.oldmeldetime = 0;
    V.bb = new double[qq];
    V.ee = new double[qq];
    V.aa = new double[pp];
    V.yy = new double[pp];

    V.aa[0] = aa0;
    switch (V.spect_type) {
        case 5:
            // AR2 spectrum with strong resonnance
            V.bb[0] = bb0;
            V.aa[1] = aa1;
            V.aa[2] = aa2;
            
            armaacf(V.p, V.aa, V.q, V.bb, 1, &rxx0, NULL);
            
            if (rxx0 != 0) {
                eev = vvar / rxx0;
            } else {
                eev = 0.0;
            }
            V.ees = sqrt(eev);
            FORLIM = V.q;
            for (i = 0; i <= FORLIM; ++i) {
                V.bb[i] *= V.ees;
            }
            armaacf(V.p, V.aa, V.q, V.bb, 1, &rxx0, NULL);
            if (rxx0 != 0) {
                eev = vvar / rxx0;
            } else {
                eev = 0.0;
            }
            yyv = vvar;
            V.ees = sqrt(eev);
            yys = sqrt(yyv);
            break;

        default:
            // all others spectral types
            mfourier_ = new tcomplex[qq];
            FORLIM = V.q;
            for (i = 0; i <= FORLIM; ++i) {
                mfourier_[i].im = 0.0;
            }
            mfourier_[0].re = sqrt(sf(0.0, V.spect_type) * V.prim_freq);
            FORLIM = V.q;
            for (i = 1; i <= FORLIM; ++i) {
                mfourier_[i].re = sqrt((sf(V.prim_freq * i / (V.q + 1), V.spect_type) +
			            sf(V.prim_freq * (V.q - i + 1) / (V.q + 1), V.spect_type)) * V.prim_freq);
            }
            ift(mfourier_, V.q + 1);
            FORLIM = V.q;
            for (i = 0; i <= FORLIM; ++i) {
                V.bb[i] = mfourier_[(i + int(floor((V.q + 1.0) / 2 + 0.5))) % (V.q + 1)].re;
            }
            delete [] mfourier_;
            double s0 = 0.0;
            FORLIM = V.q;
            for (i = 0; i <= FORLIM; ++i) {
                double TEMP = V.bb[i];
                s0 += TEMP * TEMP;
            }
            FORLIM = V.q;
            for (i = 0; i <= FORLIM; ++i) {
                V.bb[i] /= sqrt(s0);
            }
            eev = 1.0;
            V.ees = sqrt(eev);
            yyv = 0.0;
            FORLIM = V.q;
            for (i = 0; i <= FORLIM; ++i) {
                double TEMP = V.bb[i];
                yyv += TEMP * TEMP;
            }
            yys = sqrt(yyv);
            break;
    }

    // writing coefficients of the model in *.coe
    write_cof(V.out_file_basename, V);

    // Initialisierung yyv-->dtm;dsm
    double s0 = 0.0;
    for (i = -1000; i <= 1000; ++i) {
        s0 += 100 * gauss(i / 100.0);
    }
    for (int j = 0; j <= 99; ++j) {
        for (i = -1000; i <= 1000; ++i) {
            double TEMP = yys * i / 100 + V.mean + (byte)V.vosc * V.oscamp * cos(2 * M_PI * j / 100);
            s1 += gauss(i / 100.0) * fabs(sqrt(fabs(V.quer * V.quer + TEMP * TEMP)));
        }
    }
    
    V.dsm = s1 / s0 / V.dotn;
    
    V.dtm = 1 / V.dotn;

    // Initialisierung ee,yy
    if (V.p > 0) {
        // number of AR-Coefficients > 0
        int TEMP1 = pp + qq - 2;
        cc = new double[TEMP1 * TEMP1];
        TEMP1 = pp + qq - 2;
        cn = new double[TEMP1 * TEMP1];
        co = new double[pp + qq - 2];
        ryy = new double[maxi(pp - 1.0, qq - 1.0)];
        rey = new double[maxi(pp - 1.0, qq - 1.0)];
        
        armaacf(pp - 1, V.aa, qq - 1, V.bb, maxi(pp - 1.0, qq - 1.0), ryy, rey);
        
        FORLIM = maxi(pp - 1.0, qq - 1.0);
        for (i = 0; i < FORLIM; ++i) {
            ryy[i] *= V.ees * V.ees;
        }
        FORLIM = maxi(pp - 1.0, qq - 1.0);
        for (i = 0; i < FORLIM; ++i) {
            rey[i] *= V.ees * V.ees;
        }
        FORLIM = qq - 2;
        for (i = 0; i <= FORLIM; ++i) {
            int FORLIM1 = qq - 2;
            for (int j = 0; j <= FORLIM1; ++j) {
                cc[(pp + qq - 2) * i + j] = (byte)(i == j) * V.ees * V.ees;
            }
        }
        FORLIM = pp - 2;
        for (i = 0; i <= FORLIM; ++i) {
            int FORLIM1 = pp - 2;
            for (int j = 0; j <= FORLIM1; ++j) {
                cc[(pp + qq - 2) * (qq + i - 1) + qq + j - 1] = ryy[abs(i - j)];
            }
        }
        FORLIM = qq - 2;
        for (i = 0; i <= FORLIM; ++i) {
            int FORLIM1 = pp - 2;
            for (int j = 0; j <= FORLIM1; ++j) {
                cc[(pp + qq - 2) * i + qq + j - 1] = (byte)(i >= j) * rey[abs(i - j)];
            }
        }
        FORLIM = pp - 2;
        for (i = 0; i <= FORLIM; ++i) {
            int FORLIM1 = qq - 2;
            for (int j = 0; j <= FORLIM1; ++j) {
                cc[(pp + qq - 2) * (qq + i - 1) + j] = (byte)(i <= j) * rey[abs(i - j)];
            }
        }
        
        dreisqrtm_ext(cc, cn, pp + qq - 2);
        
        FORLIM = pp + qq - 3;
        for (i = 0; i <= FORLIM; ++i) {
            co[i] = gaussran();
        }
        FORLIM = qq;
        for (i = 1; i < FORLIM; ++i) {
            V.ee[qq - i] = 0.0;
            int FORLIM1 = pp + qq - 3;
            for (int j = 0; j <= FORLIM1; ++j) {
                V.ee[qq - i] += co[j] * cn[(i - 1) * (pp + qq - 2) + j];
            }
        }
        FORLIM = pp;
        for (i = 1; i < FORLIM; ++i) {
            V.yy[pp - i] = 0.0;
            int FORLIM1 = pp + qq - 3;
            for (int j = 0; j <= FORLIM1; ++j) {
                V.yy[pp - i] += co[j] * cn[(i + qq - 2) * (pp + qq - 2) + j];
            }
        }

        delete [] cc;
        delete [] cn;
        delete [] co;
        delete [] ryy;
        delete [] rey;
    } else {
        // number of AR-Coefficients == 0
        for (size_t i = 0; i <= V.q; ++i) {
            V.ee[i] = gaussran();
        }
    }

    //Start Serie
    sprintf(V.psf_NAME, "%s.ts", V.out_file_basename);
    V.psf = fopen(V.psf_NAME, "w");
    if (V.psf == NULL) {
        ERROPENF(V.psf_NAME);
    }
    
    V.gsu = 0.0;
    V.tsu = 0.0;
    V.psu = 0.0;
    V.gsc = 0;
    V.tsc = 0;
    V.psc = 0;
    V.gst = 0.0;
    V.tst = 0.0;
    V.pst = 0.0;
    V.gss = 0.0;
    V.tsx = 1.0;
    V.pstt = 0.0;
    V.psv = false;
    V.pper = 2 * M_PI * gleichran();

    // writing the data
    printf("Writing series:     ");
//     if (prim) {
//         while (V.gsc < psn) {
//             nps(&V);
//         }
//     } else {
        while (V.psc < V.psn) {
            nps(V);
        }
//    }
    
    if (V.psf != NULL) {
        fclose(V.psf);
    }
    V.psf = NULL;

    delete [] V.bb;
    delete [] V.ee;
    delete [] V.aa;
    delete [] V.yy;
}

/*! start point */
int main (int argc, char* argv []) {
    PASCAL_MAIN(argc, argv);
    _sfu_init();
    _random_init();
    cout<<"This is BM3 by Holger Nobach 07-27-2001."<<endl; 
    cout<<"C++ translation Francis Micheli (Onera) 2002-12-20."<<endl; 
    LOC_run V;
	char STR1[256];
    
    // Analysis of the command line argument
	if (P_argc < 6) {
		help();
	}
    
	strcpy(V.out_file_basename, P_argv[1]);
	if (strpos2(V.out_file_basename, ".", 1) > 0) {
		sprintf(STR1, "%.*s", strpos2(V.out_file_basename, ".", 1) - 1, V.out_file_basename);
		strcpy(V.out_file_basename, STR1);
	}
	int er = sscanf(strcpy(STR1, P_argv[2]), "%d", &V.spect_type) == 0;
	if (er > 0 || V.spect_type < 1 || V.spect_type > 5) {
		help();
	}
	er = (sscanf(strcpy(STR1, P_argv[3]), "%d", &V.bias_type) == 0);
	if (er > 0 || V.bias_type > 3) {
		help();
	}
	er = (sscanf(strcpy(STR1, P_argv[4]), "%d", &V.sig_group) == 0);
	if (er > 0 || V.sig_group > 63) {
		help();
	}
	er = (sscanf(strcpy(STR1, P_argv[5]), "%d", &V.num_sandseed) == 0);
	if (er > 0 || V.num_sandseed < 0) {
		help();
	}
	V.psn = 100000;
	if (P_argc > 6) {
		er = (sscanf(strcpy(STR1, P_argv[6]), "%d", &V.psn) == 0);
	}
	if (er > 0 || V.psn < 0) {
		help();
	}

	setrandom(r80);
	randseed = V.num_sandseed * 1000000;
	if (V.num_sandseed == 0) {
		_randomize();
	}
    
    // Computations starts here 
    init(V);
    // display result of analysis
    info(V);
    // crunch numbers
    run(V);
}

// End.
