// From input file "bm3.pas"
// $Id: bm3_tools.cpp,v 1.2 2004/09/01 12:45:28 micheli Exp $
// utilities from bm3

#include <bm3.h>
#include <common.h>

#include <cstring>
#include <cmath>
#include <iostream>
#include <ctime>
#include <sfu.h>

void
help()
{
	cout<<"Use: BM3 output stype btype group realization [samples]\n";
	cout<<"             |     |     |     |     |            |\n";
	cout<<"             |     |     |     |     |            Number of LDA samples\n";
	cout<<"             |     |     |     |     |            (1..; default: 100000)\n";
	cout<<"             |     |     |     |     Numero of realization (0,1..)\n";
	cout<<"             |     |     |     Signal group (0..63)\n";
	cout<<"             |     |     Bias type (0..3)\n";
	cout<<"             |     Spectral type (1..5)\n";
	cout<<"             Name of the files\n";
    cout<<endl;
	exit(1);
}

/*! only used for info messages */
char *
sreal(char* Result, int i)
{
    TRACE("sreal(..)");
	char s[256];

	if (i == 0) {
		return strcpy(Result, "random");
	} else {
		sprintf(s, "%d", i);
		return strcpy(Result, s);
	}
}

/*! only used for info messages*/
char *
snone(char* Result, double x)
{
    TRACE("snone(..)");
	char s[256];

	if (x == 0) {
		return strcpy(Result, "none");
	} else {
		sprintf(s, "%0.3f", x);
		return strcpy(Result, s);
	}
}

/*! only used for info messages*/
char *
sbool(char* Result, bool v)
{
    TRACE("sbool(..)");
	if (v) {
		return strcpy(Result, "yes");
	} else {
		return strcpy(Result, "no");
	}
}

void
meldung(double x, bool overwrite, LOC_run& V)
{
 	time_t newmeldetime = time(NULL);
 	if (!((newmeldetime - V.oldmeldetime > 2) || overwrite))  {
		return;
 	}
	cout<<floor(100 * x + 0.5)<<"% done."<<endl;
	V.oldmeldetime = newmeldetime;
}

void write_cof(const char* out_file_basename, const LOC_run& V)
{
    TRACE("write_cof(...)");
    // name of the arma coefficient file
    char cof_NAME[_FNSIZE];
    // arma coefficient file
    FILE* cof;
    sprintf(cof_NAME, "%s.coe", out_file_basename);
    cof = fopen(cof_NAME, "w+b");
    if (cof == NULL) {
        ERROPENF(cof_NAME);
    }
    
    fwrite(&V.prim_freq, sizeof(double), 1, cof);
    double s0 = V.p;
    fwrite(&s0, sizeof(double), 1, cof);
    s0 = V.q;
    fwrite(&s0, sizeof(double), 1, cof);

    for (size_t i = 0; i <= V.p; ++i) {
        fwrite(&V.aa[i], sizeof(double), 1, cof);
    }

    for (size_t i = 0; i <= V.q; ++i) {
        fwrite(&V.bb[i], sizeof(double), 1, cof);
    }
    fclose(cof);
}

void info(const LOC_run& V, ostream& s)
{
    char STR1[256];
	s<<"Output Name:        "<<V.out_file_basename<<'\n';
	s<<"Spectral Type:      "<<V.spect_type<<'\n';
	s<<"Bias:               Type "<<V.bias_type<<' ';
	switch (V.bias_type) {
        case 0:
            s<<"no bias\n";
            break;
        case 1:
            s<<"bias at high turbulence level (without mean vel.)\n";
            break;
        case 2:
            s<<"bias at low turbulence level (with mean vel.)\n";
            break;
        case 3:
            s<<"bias with a cross component\n";
            break;
	}
	s<<"Signal Group:       "<<V.sig_group<<'\n';
	s<<"Realization Number: "<<sreal(STR1, V.num_sandseed)<<'\n';
	s<<"Number of Samples:  "<<V.psn<<'\n';


	s<<"Mean Velocity:      "<<snone(STR1, V.mean)<<'\n';
	s<<"Velocity Variance:  "<<vvar<<'\n';
	s<<"Cross Velocity:     "<<snone(STR1, V.quer)<<'\n';
	s<<"AR-Coefficients:    "<<V.p + 1<<'\n';
	s<<"MA-Coefficients:    "<<V.q + 1<<'\n';
	s<<"Primary Frequency:  "<<V.prim_freq<<'\n';
	s<<"Particle Rate:      "<<V.dotn<<'\n';
	s<<"Noise Power:        "<<snone(STR1, V.sqrtnoise * V.sqrtnoise)<<'\n';
	s<<"Processor Delay:    "<<snone(STR1, V.del)<<'\n';
	s<<"Oscillation:        "<<sbool(STR1, V.vosc)<<'\n';
	if (V.vosc) {
		s<<" Osc. Amplitude:    "<<V.oscamp<<'\n';
		s<<" Osc. Periode:      "<<V.oscper<<'\n';
	}
	s<<"Varying Data Rate:  "<<sbool(STR1, V.vvdn) <<'\n';
	if (V.vvdn) {
		s<<" Rate Ratio:        "<<(1 / V.vdnfac - 1)<<'\n';
		s<<" Periode of Var.:   "<<V.vdnper<<'\n';
	}
	s<<"Drop Outs:          "<<sbool(STR1, V.vdro)<<'\n';
	if (V.vdro) {
		s<<" Time Ratio:        "<<(V.drofac / (1 - V.drofac))<<'\n';
		s<<" Drop out Periode:  "<<V.droper<<'\n';
	}
    s<<endl;
}
