// Output from p2c 1.21alpha-07.Dec.93, the Pascal-to-C translator
// From input file "random.pas"
// $Id: random.cpp,v 1.2 2004/09/01 12:45:28 micheli Exp $
// Author : Holger Nobach <holger.nobach@nambis.de>
// C translation : p2c and Francis Micheli <micheli@onera.fr>

//#include <p2c/p2c.h>
#include <common.h>
#include <cmath>
#include <mathe1.h>
#include <random.h>
#include <iostream>

using namespace std;

static byte sw;
long randseed;

void
setrandom(byte fw)
{
	sw = fw;
}

// ----------- Simple congruence type RNG --------------------
static union {
    // two 16 bits integers
    short tla[2];
    // one 32 bit integer
    long r;
};

double
random60()
{
    TRACE("random60()");
    tla[0]=0;
    tla[1]=randseed;
	r = r * 134775813 + 1;
	randseed = tla[0];
    return fabs(randseed / 2147483648.0);
}

double
random70()
{
    TRACE("random70()");
    tla[0]=0;
    tla[1]=randseed;
	r = static_cast<int>(r * 134775813.0 + 1);
	randseed = tla[0];
	return (randseed / 4294967296.0 + 0.5);
}

// ----------- Third RNG --------------------
//typedef byte tba[4];
//typedef long tla_[2];

// Static function used by random80
//! separates a long into 4 bytes, stored in 4 unsigned short
static void
split(long l, unsigned short* wa)
{
    TRACE("split(..)");
    union {
        long ll;
        byte lwa[4];
    };
    
    ll = l;
    wa[0] = lwa[0];
    wa[1] = lwa[1];
    wa[2] = lwa[2];
    wa[3] = lwa[3];
}

//! put together 4 bytes into a long
static void
glue(unsigned short* wa, long& l)
{
    TRACE("glue(.&)");
    union {
        long ll;
        byte lwa[4];
    };
     
	for (byte i = 0; i <= 3; ++i) {
		if (wa[i] > 255) {
            // I understand _Escape(0) as exit(0)
		    //_Escape(0);
		    ERRFATALE("glue : overflow at i = "<<i<<" with value "<<wa[i]);
		}
		lwa[i] = static_cast<byte>(wa[i]);
	}
    
    l = ll;
}

static void
mult(unsigned short w1, unsigned short w2, unsigned short& c,
	 unsigned short& m)
{
    TRACE("mult(..&&)");
	c = w1 * w2 / 256;
	m = (w1 * w2) & 255;
}

static void
addc(unsigned short w1, unsigned short w2, unsigned short co,
	 unsigned short& c, unsigned short& m)
{
    TRACE("addc(...&&)");
	c = (w1 + w2 + co) / 256;
	m = (w1 + w2 + co) & 255;
}


typedef unsigned short twa[4];
double
random80()
{
    TRACE("random80()");
	static long mm = 134775813;
	twa rsa, mma, ppa;
	twa xma[4], xca[4];
	double intpart;

	split(randseed, rsa);
	split(mm, mma);
    
	for (byte i = 0; i <= 3; ++i) {
		for (byte j = 0; j <= 3 - i; ++j) {
    		mult(rsa[i], mma[j], xca[i][j], xma[i][j]);
		}
	}
    
	for (byte i = 0; i <= 3; ++i) {
		ppa[i] = 0;
	}
    
	for (byte i = 0; i <= 3; ++i) {
		unsigned short xcc = (byte)(i == 0);
		for (byte j = 0; j <= 3 - i; ++j) {
	    	addc(ppa[i + j], xma[i][j], xcc, xcc, ppa[i + j]);
		}
		xcc = 0;
		for (byte j = 1; j <= 3 - i; ++j) {
		    addc(ppa[i + j], xca[i][j-1], xcc, xcc, ppa[i + j]);
		}
	}
	glue(ppa, randseed);
	return modf(randseed / 4294967296.0 + 0.5, &intpart);
}

// user access to the random generators

/*! use the generator set by setrandom(nn) */
double
gleichran()
{
    TRACE("gleichran()");
	double Result;
	switch (sw) {
		case r60:
    		Result = random60();
	    	break;

    	case r70:
	    	Result = random70();
		    break;

		default :
		    Result = random80();
    		break;
	}
	return Result;
}

/*! mean = 0 and standard deviation = 1 */
double
gaussran()
{
    TRACE("gaussran()");
	return (sqrt(-2. * log(1 - gleichran())) * cos(2. * M_PI * gleichran()));
}


double
expran()
{
    TRACE("expran()");
	return (-log(1. - gleichran()));
}

/*! Return the number of evenment during a unit time */
size_t
poissonran(double lam)
{
    TRACE("poissonran(.)");
	double s = 0.0;
	long n = 0;
    
	while (s < lam) {
		s += expran();
		++n;
	}
	return (n - 1);
}


void _random_init()
{
    TRACE("_random_init()");
	static int _was_initialized = 0;
	if (_was_initialized++) return;
	setrandom(r80);
}
// p2c: Note: Remember to call _random_init() in main program [215]

// End.
